Package: TreeSearch
Title: Phylogenetic Analysis with Discrete Character Data
Version: 1.7.0
Authors@R: c(
    person(
      "Martin R.", 'Smith',
      email = "martin.smith@durham.ac.uk", 
      role = c("aut", "cre", "cph"), 
      comment = c(ORCID = "0000-0001-5660-1727")
    ),
    person(
      given = "Martin", family = "Brazeau",
      role = c("cph"),
      comment = c(ORCID = "0000-0002-0650-1282")
    )
  )
License: GPL (>= 3)
Copyright: Incorporates C/C++ code from Morphy Phylogenetic Library by
        Martin Brazeau <https://github.com/mbrazeau/MorphyLib> (GPL3)
Description: Reconstruct phylogenetic trees from discrete data.
  Inapplicable character states are handled using the algorithm of Brazeau,
  Guillerme and Smith (2019) <doi:10.1093/sysbio/syy083> with the "Morphy"
  library, under equal or implied step weights.
  Contains a "shiny" user interface for interactive tree search and exploration
  of results, including character visualization, rogue taxon detection,
  tree space mapping, and cluster consensus trees (Smith 2022a, b)
  <doi:10.1093/sysbio/syab099>, <doi:10.1093/sysbio/syab100>.
  Profile Parsimony (Faith and Trueman, 2001) <doi:10.1080/10635150118627>, 
  Successive Approximations (Farris, 1969) <doi:10.2307/2412182>
  and custom optimality criteria are implemented.
URL: https://ms609.github.io/TreeSearch/ (doc),
        https://github.com/ms609/TreeSearch/ (devel)
BugReports: https://github.com/ms609/TreeSearch/issues/
Depends: R (>= 4.0)
Imports: ape (>= 5.6), cli (>= 3.0), cluster, fastmap, fastmatch (>=
        1.1.3), fs, future, PlotTools, promises, protoclust, Rcpp,
        Rdpack (>= 0.7), Rogue (> 2.0.0), shiny (>= 1.6.0), shinyjs,
        stats, stringi, TreeDist (>= 2.6.3), TreeTools (>= 1.16),
Suggests: knitr, phangorn (>= 2.2.1), Quartet, readxl, rmarkdown,
        shinytest, spelling, testthat, vdiffr (>= 1.0.0),
Config/Needs/check: callr, pkgbuild, rcmdcheck,
Config/Needs/coverage: covr, spelling
Config/Needs/memcheck: devtools
Config/Needs/metadata: codemeta
Config/Needs/revdeps: revdepcheck
Config/Needs/website: curl, igraph, pkgdown,
RdMacros: Rdpack
LinkingTo: Rcpp, TreeTools,
SystemRequirements: C++17
LazyData: true
ByteCompile: true
Encoding: UTF-8
Language: en-GB
VignetteBuilder: knitr
RoxygenNote: 7.3.2
NeedsCompilation: yes
Packaged: 2025-08-23 05:50:35 UTC; pjjg18
Author: Martin R. Smith [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-5660-1727>),
  Martin Brazeau [cph] (ORCID: <https://orcid.org/0000-0002-0650-1282>)
Maintainer: Martin R. Smith <martin.smith@durham.ac.uk>
Repository: CRAN
Date/Publication: 2025-08-23 06:10:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-25 21:04:24 UTC; unix
Archs: TreeSearch.so.dSYM
