# Generated by roxygen2: do not edit by hand

S3method(MaximumLength,character)
S3method(MaximumLength,numeric)
S3method(MaximumLength,phyDat)
S3method(MinimumLength,character)
S3method(MinimumLength,numeric)
S3method(MinimumLength,phyDat)
S3method(PlotCharacter,list)
S3method(PlotCharacter,multiPhylo)
S3method(PlotCharacter,phylo)
S3method(SPRMoves,matrix)
S3method(SPRMoves,phylo)
S3method(TBRMoves,matrix)
S3method(TBRMoves,phylo)
S3method(TreeLength,"NULL")
S3method(TreeLength,list)
S3method(TreeLength,multiPhylo)
S3method(TreeLength,numeric)
S3method(TreeLength,phylo)
S3method(summary,morphyPtr)
export(.NonDuplicateRoot)
export(.UniqueExceptHits)
export(AdditionTree)
export(C_MorphyLength)
export(Carter1)
export(CharacterLength)
export(ClusterStrings)
export(ClusteringConcordance)
export(ConcordantInfo)
export(ConcordantInformation)
export(Consistency)
export(DoNothing)
export(EasyTrees)
export(EasyTreesy)
export(EdgeListSearch)
export(EmptyPhyDat)
export(Evaluate)
export(ExpectedLength)
export(Fitch)
export(FitchSteps)
export(GapHandler)
export(GetMorphyLength)
export(IWScore)
export(IWTreeSearch)
export(JackLabels)
export(Jackknife)
export(LengthAdded)
export(Log2Carter1)
export(LogCarter1)
export(MaximizeParsimony)
export(MaximumLength)
export(MinimumLength)
export(MorphyBootstrap)
export(MorphyErrorCheck)
export(MorphyLength)
export(MorphyTreeLength)
export(MorphyWeights)
export(MostContradictedFreq)
export(MultiRatchet)
export(MutualClusteringConcordance)
export(NNI)
export(NNISwap)
export(PhyDat2Morphy)
export(PhylogeneticConcordance)
export(PlotCharacter)
export(PolEscapa)
export(PrepareDataIW)
export(PrepareDataProfile)
export(PresCont)
export(QuartetConcordance)
export(QuartetResolution)
export(RandomMorphyTree)
export(RandomTreeScore)
export(Ratchet)
export(RatchetConsensus)
export(RearrangeEdges)
export(Resample)
export(RootedNNI)
export(RootedNNISwap)
export(RootedSPR)
export(RootedSPRSwap)
export(RootedTBR)
export(RootedTBRSwap)
export(SPR)
export(SPRMoves)
export(SPRSwap)
export(SPRWarning)
export(SetMorphyWeights)
export(SharedPhylogeneticConcordance)
export(SingleCharMorphy)
export(StepInformation)
export(StopUnlessBifurcating)
export(Suboptimality)
export(SuccessiveApproximations)
export(SuccessiveWeights)
export(TBR)
export(TBRMoves)
export(TBRSwap)
export(TBRWarning)
export(TaxonInfluence)
export(TreeLength)
export(TreeSearch)
export(UnloadMorphy)
export(WhenFirstHit)
export(WithOneExtraStep)
export(cNNI)
export(cSPR)
export(is.morphyPtr)
export(mpl_apply_tipdata)
export(mpl_attach_rawdata)
export(mpl_attach_symbols)
export(mpl_delete_Morphy)
export(mpl_first_down_recon)
export(mpl_first_up_recon)
export(mpl_get_charac_weight)
export(mpl_get_gaphandl)
export(mpl_get_num_charac)
export(mpl_get_num_internal_nodes)
export(mpl_get_numtaxa)
export(mpl_get_symbols)
export(mpl_init_Morphy)
export(mpl_new_Morphy)
export(mpl_second_down_recon)
export(mpl_second_up_recon)
export(mpl_set_charac_weight)
export(mpl_set_gaphandl)
export(mpl_set_num_internal_nodes)
export(mpl_set_parsim_t)
export(mpl_translate_error)
export(mpl_update_lower_root)
export(mpl_update_tip)
importFrom(PlotTools,SpectrumLegend)
importFrom(Rcpp,compileAttributes)
importFrom(Rdpack,reprompt)
importFrom(Rogue,ColByStability)
importFrom(TreeDist,ClusteringEntropy)
importFrom(TreeDist,ClusteringInfo)
importFrom(TreeDist,ClusteringInfoDistance)
importFrom(TreeDist,Entropy)
importFrom(TreeDist,MutualClusteringInfo)
importFrom(TreeDist,SharedPhylogeneticInfo)
importFrom(TreeTools,AddTipEverywhere)
importFrom(TreeTools,AddUnconstrained)
importFrom(TreeTools,CharacterInformation)
importFrom(TreeTools,CladisticInfo)
importFrom(TreeTools,CompatibleSplits)
importFrom(TreeTools,Consensus)
importFrom(TreeTools,ConstrainedNJ)
importFrom(TreeTools,DescendantEdges)
importFrom(TreeTools,DescendantTips)
importFrom(TreeTools,DoubleFactorial)
importFrom(TreeTools,DropTip)
importFrom(TreeTools,EdgeAncestry)
importFrom(TreeTools,ImposeConstraint)
importFrom(TreeTools,KeepTip)
importFrom(TreeTools,Log2DoubleFactorial)
importFrom(TreeTools,Log2Unrooted)
importFrom(TreeTools,Log2UnrootedMult)
importFrom(TreeTools,LogDoubleFactorial)
importFrom(TreeTools,MakeTreeBinary)
importFrom(TreeTools,MatrixToPhyDat)
importFrom(TreeTools,NRooted)
importFrom(TreeTools,NSplits)
importFrom(TreeTools,NTip)
importFrom(TreeTools,NUnrooted)
importFrom(TreeTools,NUnrootedMult)
importFrom(TreeTools,NexusTokens)
importFrom(TreeTools,PectinateTree)
importFrom(TreeTools,PhyDatToMatrix)
importFrom(TreeTools,PhyToString)
importFrom(TreeTools,PolarizeSplits)
importFrom(TreeTools,Postorder)
importFrom(TreeTools,PostorderOrder)
importFrom(TreeTools,Preorder)
importFrom(TreeTools,RandomTree)
importFrom(TreeTools,Renumber)
importFrom(TreeTools,RenumberEdges)
importFrom(TreeTools,RenumberTips)
importFrom(TreeTools,RenumberTree)
importFrom(TreeTools,RootTree)
importFrom(TreeTools,SampleOne)
importFrom(TreeTools,SplitConflicts)
importFrom(TreeTools,SplitFrequency)
importFrom(TreeTools,StringToPhyDat)
importFrom(TreeTools,Subsplit)
importFrom(TreeTools,SupportColour)
importFrom(TreeTools,TipLabels)
importFrom(TreeTools,TreeIsRooted)
importFrom(TreeTools,as.Splits)
importFrom(TreeTools,as.multiPhylo)
importFrom(ape,consensus)
importFrom(ape,keep.tip)
importFrom(ape,nodelabels)
importFrom(ape,plot.phylo)
importFrom(ape,read.nexus)
importFrom(ape,root)
importFrom(ape,write.nexus)
importFrom(cli,cli_alert)
importFrom(cli,cli_alert_danger)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(cli,cli_alert_warning)
importFrom(cli,cli_h1)
importFrom(cli,cli_progress_bar)
importFrom(cli,cli_progress_done)
importFrom(cli,cli_progress_update)
importFrom(cluster,pam)
importFrom(cluster,silhouette)
importFrom(fastmap,fastmap)
importFrom(fastmatch,"%fin%")
importFrom(fastmatch,fmatch)
importFrom(fs,path_sanitize)
importFrom(future,future)
importFrom(graphics,par)
importFrom(promises,future_promise)
importFrom(protoclust,protoclust)
importFrom(shiny,runApp)
importFrom(shinyjs,useShinyjs)
importFrom(stats,as.dist)
importFrom(stats,cutree)
importFrom(stats,median)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(stats,weighted.mean)
importFrom(stringi,stri_paste)
importFrom(utils,adist)
importFrom(utils,combn)
useDynLib(TreeSearch, .registration = TRUE)
